"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypedDataTable = exports.TypedDocString = exports.TypedTableCell = exports.TypedTableRow = exports.TypedExamples = exports.TypedScenario = exports.TypedStep = exports.TypedBackground = exports.TypedRuleChild = exports.TypedRule = exports.TypedFeatureChild = exports.TypedComment = exports.TypedTag = exports.TypedFeature = exports.TypedGherkinDocument = exports.TypedGherkinNodeWithLocation = exports.isHasChild = exports.isHasChildren = exports.TypedGherkinNode = exports.isWithLocation = void 0;
function isWithLocation(node) {
    return typeof node === 'object' && node !== null && 'location' in node;
}
exports.isWithLocation = isWithLocation;
function reEscapeTableCell(str) {
    const out = str
        // espace all pipes that have been unescaped
        .replace('|', '\\|')
        // espace all backslashes that have been unescaped
        .replace(/\\/g, '\\\\');
    if (OPTIONS.escapeBackslashes) {
        return out;
    }
    // replace all escaped backslashes, except those that are followed by a newline or a pipe
    return out.replace(/(\\\\)(?![n\n\\])/g, '\\');
}
/**
 * Set the max size of each column in the table
 */
function generateColumnSizes(tableRows) {
    return tableRows.reduce((acc, row) => {
        row.cells.forEach((cell, index) => {
            if (!acc[index]) {
                acc[index] = 0;
            }
            acc[index] = Math.max(acc[index], reEscapeTableCell(cell.value).length);
        });
        return acc;
    }, []);
}
class TypedGherkinNode {
    constructor(originalNode) { }
}
exports.TypedGherkinNode = TypedGherkinNode;
function isHasChildren(node) {
    return typeof node === 'object' && node !== null && 'children' in node;
}
exports.isHasChildren = isHasChildren;
function isHasChild(node) {
    return typeof node === 'object' && node !== null && 'child' in node;
}
exports.isHasChild = isHasChild;
class TypedGherkinNodeWithLocation extends TypedGherkinNode {
    constructor(originalNode) {
        super(originalNode);
        this.location = originalNode.location;
    }
}
exports.TypedGherkinNodeWithLocation = TypedGherkinNodeWithLocation;
/**
 * An options object that will be set for a document
 * It is stored as a singleton that will be reset on each document to avoid passing it around for now.
 * A better option may to have access to the options on every classes.
 */
let OPTIONS;
class TypedGherkinDocument extends TypedGherkinNode {
    constructor(originalNode, options) {
        super(originalNode);
        OPTIONS = options;
        this.uri = originalNode.uri;
        this.feature = originalNode.feature
            ? new TypedFeature(originalNode.feature)
            : undefined;
        this.comments = originalNode.comments.map((c) => new TypedComment(c));
    }
    get child() {
        return this.feature;
    }
}
exports.TypedGherkinDocument = TypedGherkinDocument;
class TypedFeature extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.tags = originalNode.tags.map((t) => new TypedTag(t));
        this.language = originalNode.language;
        this.keyword = originalNode.keyword;
        this.name = originalNode.name;
        this.description = originalNode.description;
        this.children = originalNode.children.map((c) => new TypedFeatureChild(c));
    }
}
exports.TypedFeature = TypedFeature;
class TypedTag extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.name = originalNode.name;
        this.id = originalNode.id;
    }
}
exports.TypedTag = TypedTag;
class TypedComment extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.text = originalNode.text;
    }
    get value() {
        return this.text.trim();
    }
}
exports.TypedComment = TypedComment;
class TypedFeatureChild extends TypedGherkinNode {
    constructor(originalNode) {
        super(originalNode);
        this.rule = originalNode.rule
            ? new TypedRule(originalNode.rule)
            : undefined;
        this.background = originalNode.background
            ? new TypedBackground(originalNode.background)
            : undefined;
        this.scenario = originalNode.scenario
            ? new TypedScenario(originalNode.scenario)
            : undefined;
    }
    get child() {
        return this.rule || this.background || this.scenario;
    }
}
exports.TypedFeatureChild = TypedFeatureChild;
class TypedRule extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.tags = originalNode.tags.map((t) => new TypedTag(t));
        this.keyword = originalNode.keyword;
        this.name = originalNode.name;
        this.description = originalNode.description;
        this.children = originalNode.children.map((c) => new TypedRuleChild(c));
        this.id = originalNode.id;
    }
}
exports.TypedRule = TypedRule;
class TypedRuleChild extends TypedGherkinNode {
    constructor(originalNode) {
        super(originalNode);
        this.background = originalNode.background
            ? new TypedBackground(originalNode.background)
            : undefined;
        this.scenario = originalNode.scenario
            ? new TypedScenario(originalNode.scenario)
            : undefined;
    }
    get child() {
        return this.scenario;
    }
}
exports.TypedRuleChild = TypedRuleChild;
class TypedBackground extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.keyword = originalNode.keyword;
        this.name = originalNode.name;
        this.description = originalNode.description;
        this.steps = originalNode.steps.map((s) => new TypedStep(s));
        this.id = originalNode.id;
    }
    get children() {
        return this.steps;
    }
}
exports.TypedBackground = TypedBackground;
class TypedStep extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.keyword = originalNode.keyword;
        this.keywordType = originalNode.keywordType;
        this.text = originalNode.text;
        this.docString = originalNode.docString
            ? new TypedDocString(originalNode.docString)
            : undefined;
        this.dataTable = originalNode.dataTable
            ? new TypedDataTable(originalNode.dataTable)
            : undefined;
        this.id = originalNode.id;
    }
    get children() {
        return [this.docString, this.dataTable].filter((c) => typeof c !== 'undefined');
    }
}
exports.TypedStep = TypedStep;
class TypedScenario extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.location = originalNode.location;
        this.tags = originalNode.tags.map((t) => new TypedTag(t));
        this.keyword = originalNode.keyword;
        this.name = originalNode.name;
        this.description = originalNode.description;
        this.steps = originalNode.steps.map((s) => new TypedStep(s));
        this.examples = originalNode.examples.map((e) => new TypedExamples(e));
        this.id = originalNode.id;
    }
    get children() {
        return [...this.steps, ...this.examples, ...this.tags];
    }
}
exports.TypedScenario = TypedScenario;
class TypedExamples extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        var _a;
        super(originalNode);
        const rows = [
            originalNode.tableHeader,
            ...((_a = originalNode.tableBody) !== null && _a !== void 0 ? _a : []),
        ].filter((row) => typeof row !== 'undefined');
        const columnSizes = generateColumnSizes(rows);
        this.tags = originalNode.tags.map((t) => new TypedTag(t));
        this.keyword = originalNode.keyword;
        this.name = originalNode.name;
        this.description = originalNode.description;
        this.tableHeader = originalNode.tableHeader
            ? new TypedTableRow(originalNode.tableHeader, columnSizes)
            : undefined;
        this.tableBody = originalNode.tableBody.map((r) => new TypedTableRow(r, columnSizes));
        this.id = originalNode.id;
    }
    get children() {
        return [
            ...(this.tableHeader ? [this.tableHeader] : []),
            ...this.tableBody,
            ...this.tags,
        ];
    }
}
exports.TypedExamples = TypedExamples;
class TypedTableRow extends TypedGherkinNodeWithLocation {
    constructor(originalNode, columnSizes) {
        super(originalNode);
        this.cells = originalNode.cells.map((c, index) => new TypedTableCell(c, columnSizes[index]));
        this.id = originalNode.id;
        this.columnSizes = columnSizes;
    }
    get children() {
        return this.cells;
    }
}
exports.TypedTableRow = TypedTableRow;
class TypedTableCell extends TypedGherkinNodeWithLocation {
    constructor(originalNode, displaySize) {
        super(originalNode);
        this.value = reEscapeTableCell(originalNode.value);
        this.displaySize = Math.max(displaySize !== null && displaySize !== void 0 ? displaySize : 0, this.value.length);
    }
}
exports.TypedTableCell = TypedTableCell;
class TypedDocString extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        this.mediaType = originalNode.mediaType;
        this.content = originalNode.content;
        this.delimiter = originalNode.delimiter;
    }
}
exports.TypedDocString = TypedDocString;
class TypedDataTable extends TypedGherkinNodeWithLocation {
    constructor(originalNode) {
        super(originalNode);
        const columnSizes = generateColumnSizes(originalNode.rows);
        this.rows = originalNode.rows.map((r) => new TypedTableRow(r, columnSizes));
    }
    get children() {
        return this.rows;
    }
}
exports.TypedDataTable = TypedDataTable;
